# cython: profile=True

from chat_trie cimport *

cdef class BaseTrie:
    """
    Base HAT-Trie wrapper.
    """

    cdef hattrie_t* _trie

    def __cinit__(self):
        self._trie = hattrie_create()

    def __dealloc__(self):
        if self._trie:
            hattrie_free(self._trie)


    def __getitem__(self, bytes key):
        return self._getitem(key)

    def __setitem__(self, bytes key, int value):
        self._setitem(key, value)

    def __contains__(self, bytes key):
        return self._contains(key)

    def __len__(self):
        return (<hattrie_t_*> self._trie).m

    def setdefault(self, bytes key, int value):
        return self._setdefault(key, value)

    def keys(self):
        return list(self.iterkeys())

    def iterkeys(self):
        cdef:
            hattrie_iter_t* it = hattrie_iter_begin(self._trie, 0)
            char* c_key
            size_t val
            size_t length
            bytes py_str

        try:
            while not hattrie_iter_finished(it):
                c_key = hattrie_iter_key(it, &length)
                py_str = c_key[:length]
                yield py_str
                hattrie_iter_next(it)

        finally:
            hattrie_iter_free(it)


    cdef int _getitem(self, char* key) except -1:
        cdef value_t* value_ptr = hattrie_tryget(self._trie, key, len(key))
        if value_ptr == NULL:
            raise KeyError(key)
        return value_ptr[0]

    cdef void _setitem(self, char* key, value_t value):
        hattrie_get(self._trie, key, len(key))[0] = value

    cdef int _setdefault(self, char* key, value_t value):
        cdef value_t* value_ptr = hattrie_tryget(self._trie, key, len(key))
        if value_ptr == NULL:
            self._setitem(key, value)
            return value
        return value_ptr[0]

    cdef bint _contains(self, char* key):
        cdef value_t* value_ptr = hattrie_tryget(self._trie, key, len(key))
        return value_ptr != NULL


cdef class Trie(BaseTrie):
    """
    HAT-Trie with unicode support.

    XXX: Internal encoding is hardcoded as UTF8. This is the fastest
    encoding that can handle all unicode symbols and doesn't have
    zero bytes.

    This may seem sub-optimal because it is multibyte encoding;
    single-byte language-specific encoding (such as cp1251)
    seems to be faster. But this is not the case because:

    1) the bottleneck of this wrapper is string encoding, not trie traversal;
    2) python's unicode encoding utilities are optimized for utf8;
    3) users will have to select language-specific encoding for the trie;
    4) non-hardcoded encoding causes extra overhead and prevents cython
       optimizations.

    That's why hardcoded utf8 is up to 9 times faster than configurable cp1251.

    XXX: char-walking utilities may become tricky with multibyte
    internal encoding.
    """

    def __getitem__(self, unicode key):
        cdef bytes bkey = key.encode('utf8')
        return self._getitem(bkey)

    def __contains__(self, unicode key):
        cdef bytes bkey = key.encode('utf8')
        return self._contains(bkey)

    def __setitem__(self, unicode key, int value):
        cdef bytes bkey = key.encode('utf8')
        self._setitem(bkey, value)

    def setdefault(self, unicode key, int value):
        cdef bytes bkey = key.encode('utf8')
        return self._setdefault(bkey, value)

    def keys(self):
        return [key.decode('utf8') for key in self.iterkeys()]



cdef extern from "../hat-trie/src/hat-trie.h":

    ctypedef int value_t
    ctypedef int size_t

    ctypedef struct hattrie_t:
        pass

    hattrie_t* hattrie_create ()                 # Create an empty hat-trie.
    void       hattrie_free   (hattrie_t*)       # Free all memory used by a trie.
    hattrie_t* hattrie_dup    (hattrie_t*)       # Duplicate an existing trie.
    void       hattrie_clear  (hattrie_t*)       # Remove all entries.


    # Find the given key in the trie, inserting it if it does not exist, and
    # returning a pointer to it's key.
    # This pointer is not guaranteed to be valid after additional calls to
    # hattrie_get, hattrie_del, hattrie_clear, or other functions that
    # modifies the trie.
    value_t* hattrie_get (hattrie_t*, char* key, size_t len)

    # Find a given key in the table, returning a NULL pointer if it does not exist.
    value_t* hattrie_tryget (hattrie_t*, char* key, size_t len)

    ctypedef struct hattrie_iter_t:
        pass

    hattrie_iter_t* hattrie_iter_begin     (hattrie_t*, bint sorted)
    void            hattrie_iter_next      (hattrie_iter_t*)
    bint            hattrie_iter_finished  (hattrie_iter_t*)
    void            hattrie_iter_free      (hattrie_iter_t*)
    char*           hattrie_iter_key       (hattrie_iter_t*, size_t* len)
    value_t*        hattrie_iter_val       (hattrie_iter_t*)

cdef struct hattrie_t_:
    void* root
    size_t m      # number of stored keys



#! /usr/bin/env python
import os
from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

HATTRIE_DIR = 'hat-trie/src'
HATTRIE_FILE_NAMES = ['ahtable.c', 'hat-trie.c', 'misc.c', 'murmurhash3.c']
HATTRIE_FILES = [os.path.join(HATTRIE_DIR, name) for name in HATTRIE_FILE_NAMES]

setup(
    name="hat-trie",
    version="0.1",
    description="HAT-Trie for Python",
    long_description = open('README.rst').read() + "\n\n" + open('CHANGES.rst').read(),
    author='Mikhail Korobov',
    author_email='kmike84@gmail.com',
    url='https://github.com/kmike/hat-trie/',
    #packages = ['hat_trie'],
    cmdclass = {'build_ext': build_ext},

    ext_modules = [
        Extension(
            "hat_trie",
            ['src/hat_trie.pyx', 'src/chat_trie.pxd'] + HATTRIE_FILES,
            #['src/datrie.c', 'src/cdatrie.c', 'src/stdio_ext.c'] + HATTRIE_FILES,
            include_dirs=['hat-trie/src'],
        )
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing :: Linguistic',
    ],
)
